-------------------------------------------------------------------------------
-- LoreBooks v4.2
-------------------------------------------------------------------------------
--
-- Copyright (c) 2014, 2015 Ales Machat (Garkin)
--
-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.
--
-------------------------------------------------------------------------------
--
-- DISCLAIMER:
--
-- This Add-on is not created by, affiliated with or sponsored by ZeniMax
-- Media Inc. or its affiliates. The Elder Scrolls® and related logos are
-- registered trademarks or trademarks of ZeniMax Media Inc. in the United
-- States and/or other countries. All rights reserved.
--
-- You can read the full terms at:
-- https://account.elderscrollsonline.com/add-on-terms
--
-------------------------------------------------------------------------------

--Libraries--------------------------------------------------------------------
local LAM = LibStub("LibAddonMenu-2.0")
local LMP = LibStub("LibMapPins-1.0")
local GPS = LibStub("LibGPS2")
local Postmail = {}
local lastInteractionActionWas
local lastReticleUpdateWas = 0

--Local constants -------------------------------------------------------------
local ADDON_VERSION = "4.2"
local PINS_UNKNOWN = "LBooksMapPin_unknown"
local PINS_COLLECTED = "LBooksMapPin_collected"
local PINS_EIDETIC = "LBooksMapPin_eidetic"
--local PINS_EIDETIC_COLLECTED = "LBooksMapPin_eideticCollected"
local PINS_COMPASS = "LBooksCompassPin_unknown"
local PINS_COMPASS_EIDETIC = "LBooksCompassPin_eidetic"

local MISSING_TEXTURE = "/esoui/art/icons/icon_missing.dds"
local PLACEHOLDER_TEXTURE = "/esoui/art/icons/lore_book4_detail1_color2.dds"

--Local variables -------------------------------------------------------------
local lang = GetCVar("Language.2")
local updatePins = {}
local loreBoooksLibrary = {}
local totalCurrentlyCollected = 0
local updating = false
local mapIsShowing
local db							--user settings
local svData					--collected books
local newBooks = {}			--books which are not listed in LoreBooksData.lua
local defaults = {			--default settings for saved variables
	compassMaxDistance = 0.04,
	pinTexture = {
		type = 1,
		size = 26,
		level = 40,
	},
	filters = {
		[PINS_COMPASS_EIDETIC] = false,
		[PINS_COMPASS] = true,
		[PINS_UNKNOWN] = true,
		[PINS_COLLECTED] = false,
		[PINS_EIDETIC] = false,
		--[PINS_EIDETIC_COLLECTED] = false,
	},
	shareData = false,
	postmailData = "",
	postmailFirstInsert = GetTimeStamp(),
	booksFound = {},
	nbBooksFound = 0,
	booksCollected = {},
	unlockEidetic = false,
}

local INFORMATION_TOOLTIP
local loreLibraryReportKeybind
local reportShown
local copyReport

-- Local functions ------------------------------------------------------------
--prints message to chat
local function MyPrint(...)
	CHAT_SYSTEM:AddMessage(...)
end

--checks if normalized point is valid
local function InvalidPoint(x, y)
	return (x < 0 or x > 1 or y < 0 or y > 1)
end

local function GetEpsilon(zone, subzone)
	--Epsilon value should be bigger for small maps and smaller for big maps.
	-- As I have to determine epsilon even when it's not possible to check
	-- mapsize directly, I must find some kind of workaround...
	local epsilon = 0.007

	if subzone == "ava_whole" then							--Cyrodiil zone map
		epsilon = 0.003
	elseif zone == "cyrodiil" then							--all other subzones in Cyrodiil
		epsilon = 0.010
	elseif (subzone):find(zone) then							--zones
		epsilon = 0.005
	elseif zone == "guildmaps" or zone == "main" then	--Guild Maps, Main quest maps
		epsilon = 0.012
	end

	return epsilon
end

local function UnknownCoords(dataTable, x, y, epsilon)
	if not dataTable or dataTable == {} then
		return true
	end

	epsilon = epsilon or 0.005

	for i, data in ipairs(dataTable) do
		if zo_floatsAreEqual(data[1], x, epsilon) and zo_floatsAreEqual(data[2], y, epsilon) then
			return false
		end
	end

	return true
end

-- Pins -----------------------------------------------------------------------
local pinTexturesList = {
	[1] = "Shalidor's Library icons",
	[2] = "Book icon set 1",
	[3] = "Book icon set 2",
	[4] = "Esohead's icons (Rushmik)",
}

local pinTextures = {
	--[index] = { known_book_texture, unknown_book_texture },
	[1] = { "EsoUI/Art/Icons/lore_book4_detail4_color5.dds", "EsoUI/Art/Icons/lore_book4_detail4_color5.dds" },
	[2] = { "LoreBooks/Icons/book1.dds", "LoreBooks/Icons/book1-invert.dds" },
	[3] = { "LoreBooks/Icons/book2.dds", "LoreBooks/Icons/book2-invert.dds" },
	[4] = { "LoreBooks/Icons/book3.dds", "LoreBooks/Icons/book3-invert.dds" },
}

local function GetPinTexture(self)
	local _, texture, known = GetLoreBookInfo(1, self.m_PinTag[3], self.m_PinTag[4])
	local textureType = db.pinTexture.type
	if texture == MISSING_TEXTURE then texture = PLACEHOLDER_TEXTURE end
	return (pinTexturesList[textureType] == "Shalidor's Library icons") and texture or pinTextures[textureType][known and 1 or 2]
end

local function GetPinTextureEidetic(self)
	local _, texture, known = GetLoreBookInfo(2, self.m_PinTag.c, self.m_PinTag.b)
	local textureType = db.pinTexture.type
	if texture == MISSING_TEXTURE then texture = PLACEHOLDER_TEXTURE end
	return (pinTexturesList[textureType] == "Shalidor's Library icons") and texture or pinTextures[textureType][known and 1 or 2]
end

local function IsPinGrayscale()
	return pinTexturesList[db.pinTexture.type] == "Shalidor's Library icons"
end

--tooltip creator
local pinTooltipCreator = {}
pinTooltipCreator.tooltip = 1 --TOOLTIP_MODE.INFORMATION
pinTooltipCreator.creator = function(pin)
	
	local pinTag = pin.m_PinTag
	local title, icon, known = GetLoreBookInfo(1, pinTag[3], pinTag[4])
	local collection = GetLoreCollectionInfo(1, pinTag[3])
	local moreinfo = {}
	if icon == MISSING_TEXTURE then icon = PLACEHOLDER_TEXTURE end
	
	if pinTag[5] then
		table.insert(moreinfo, "[" .. GetString("LBOOKS_MOREINFO", pinTag[5]) .. "]")
	end
	if known then
		table.insert(moreinfo, "[" .. GetString(LBOOKS_KNOWN) .. "]")
	end
	
	if IsInGamepadPreferredMode() then
		INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, zo_strformat(collection), INFORMATION_TOOLTIP.tooltip:GetStyle("mapTitle"))
		INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, icon, title, {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_3})
		if #moreinfo > 0 then
			INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, table.concat(moreinfo, " / "), INFORMATION_TOOLTIP.tooltip:GetStyle("worldMapTooltip"))
		end
	else
		INFORMATION_TOOLTIP:AddLine(zo_strformat(collection), "ZoFontGameOutline", ZO_SELECTED_TEXT:UnpackRGB())
		ZO_Tooltip_AddDivider(INFORMATION_TOOLTIP)
		INFORMATION_TOOLTIP:AddLine(zo_iconTextFormat(icon, 32, 32, title), "", ZO_HIGHLIGHT_TEXT:UnpackRGB())
		if #moreinfo > 0 then
			INFORMATION_TOOLTIP:AddLine(table.concat(moreinfo, " / "), "", ZO_TOOLTIP_DEFAULT_COLOR:UnpackRGB())
		end
	end
	
end

--tooltip creator
local pinTooltipCreatorEidetic = {}
pinTooltipCreatorEidetic.tooltip = 1 --TOOLTIP_MODE.INFORMATION
pinTooltipCreatorEidetic.creator = function(pin)
	
	local pinTag = pin.m_PinTag
	local title, icon, known = GetLoreBookInfo(2, pinTag.c, pinTag.b)
	local collection = GetLoreCollectionInfo(2, pinTag.c)
	if icon == MISSING_TEXTURE then icon = PLACEHOLDER_TEXTURE end
	
	if IsInGamepadPreferredMode() then
		
		INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, zo_strformat(collection), INFORMATION_TOOLTIP.tooltip:GetStyle("mapTitle"))
		INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, title, {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_3})
		
		if pinTag.q then
			if type(pinTag.q) == "boolean" then
				INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, GetString(LBOOKS_LINK_TO_UNKN_QUEST), {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
			elseif type(pinTag.q) == "table" then
				INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, GetString(LBOOKS_QUEST_BOOK), {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
				INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, pinTag.q[lang] or pinTag.q["en"], {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
				
				local questDetails
				if pinTag.qt then
					questDetails = zo_strformat(GetString("LBOOKS_SPECIAL_QUEST"), pinTag.qt)
				else
					questDetails = zo_strformat(GetString(LBOOKS_QUEST_IN_ZONE), zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetMapNameByIndex(pinTag.qm)))
				end
				
				INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, questDetails, {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
			end
			if pinTag.u then
				INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, GetString(LBOOKS_BOOK_DISAPEAR), {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
			end
		end
		
		if pinTag.d then
		
			local dungeonDifficulty = ""
			if pinTag.v ~= nil then -- can be false
				if pinTag.v then
					dungeonDifficulty = zo_strformat("(<<1>>)", GetString(SI_DUNGEONDIFFICULTY2))
				else
					dungeonDifficulty = zo_strformat("(<<1>>)", GetString(SI_DUNGEONDIFFICULTY1))
				end
			end
		
			if GetZoneNameByIndex(pinTag.z) == GetMapNameByIndex(pinTag.m) then
				INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, zo_strformat("[<<1>>]", GetString(SI_QUESTTYPE5)), {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
			else
				INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, zo_strformat("[<<1>>] <<2>>", zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetZoneNameByIndex(pinTag.z)), dungeonDifficulty), {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
			end
			
		end
		
		if pinTag.r then
			INFORMATION_TOOLTIP:LayoutIconStringLine(INFORMATION_TOOLTIP.tooltip, nil, GetString(LBOOKS_RANDOM_POSITION), {fontSize = 27, fontColorField = GAMEPAD_TOOLTIP_COLOR_GENERAL_COLOR_2})
		end
		
	else
		
		INFORMATION_TOOLTIP:AddLine(zo_strformat(collection), "ZoFontGameOutline", ZO_SELECTED_TEXT:UnpackRGB())
		ZO_Tooltip_AddDivider(INFORMATION_TOOLTIP)
		
		local bookColor = ZO_HIGHLIGHT_TEXT
		if known then
			bookColor = ZO_SUCCEEDED_TEXT
		end
		
		INFORMATION_TOOLTIP:AddLine(zo_iconTextFormat(icon, 32, 32, title), "", bookColor:UnpackRGB())
		
		if pinTag.q then
			if type(pinTag.q) == "boolean" then
				INFORMATION_TOOLTIP:AddLine(GetString(LBOOKS_LINK_TO_UNKN_QUEST), "", ZO_SELECTED_TEXT:UnpackRGB())
			elseif type(pinTag.q) == "table" then
				INFORMATION_TOOLTIP:AddLine(GetString(LBOOKS_QUEST_BOOK), "", ZO_SELECTED_TEXT:UnpackRGB())
				INFORMATION_TOOLTIP:AddLine(zo_strformat("[<<1>>]", pinTag.q[lang] or pinTag.q["en"]), "", ZO_SELECTED_TEXT:UnpackRGB())
				
				local questDetails
				if pinTag.qt then
					questDetails = zo_strformat(GetString("LBOOKS_SPECIAL_QUEST"), pinTag.qt)
				else
					questDetails = zo_strformat(GetString(LBOOKS_QUEST_IN_ZONE), zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetMapNameByIndex(pinTag.qm)))
				end
				
				INFORMATION_TOOLTIP:AddLine(questDetails, "", ZO_HIGHLIGHT_TEXT:UnpackRGB())
			end
			if pinTag.u then
				INFORMATION_TOOLTIP:AddLine(GetString(LBOOKS_BOOK_DISAPEAR), "", ZO_SELECTED_TEXT:UnpackRGB())
			end
		end		
		
		if pinTag.d then
		
			local dungeonDifficulty = ""
			if pinTag.v ~= nil then -- can be false
				if pinTag.v then
					dungeonDifficulty = zo_strformat("(<<1>>)", GetString(SI_DUNGEONDIFFICULTY2))
				else
					dungeonDifficulty = zo_strformat("(<<1>>)", GetString(SI_DUNGEONDIFFICULTY1))
				end
			end
		
			if GetZoneNameByIndex(pinTag.z) == GetMapNameByIndex(pinTag.m) then
				INFORMATION_TOOLTIP:AddLine(zo_strformat("[<<1>>]", GetString(SI_QUESTTYPE5)), "", ZO_SELECTED_TEXT:UnpackRGB())
			else
				INFORMATION_TOOLTIP:AddLine(zo_strformat("[<<1>>] <<2>>",  zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetZoneNameByIndex(pinTag.z)), dungeonDifficulty), "", ZO_SELECTED_TEXT:UnpackRGB())
			end
		end
		
		if pinTag.r then
			INFORMATION_TOOLTIP:AddLine(GetString(LBOOKS_RANDOM_POSITION))
		end
		
	end
	
end

local function CreatePins()
	
	if (updatePins[PINS_COLLECTED] and LMP:IsEnabled(PINS_COLLECTED)) or (updatePins[PINS_UNKNOWN] and LMP:IsEnabled(PINS_UNKNOWN)) or (updatePins[PINS_COMPASS] and db.filters[PINS_COMPASS]) then
		local zone, subzone = LMP:GetZoneAndSubzone()
		local lorebooks = LoreBooks_GetLocalData(zone, subzone)
		if lorebooks then
			for _, pinData in ipairs(lorebooks) do	
				local _, _, known = GetLoreBookInfo(1, pinData[3], pinData[4])
				if known and updatePins[PINS_COLLECTED] and LMP:IsEnabled(PINS_COLLECTED) then
					LMP:CreatePin(PINS_COLLECTED, pinData, pinData[1], pinData[2])
				elseif not known then
					if updatePins[PINS_UNKNOWN] and LMP:IsEnabled(PINS_UNKNOWN) then
						LMP:CreatePin(PINS_UNKNOWN, pinData, pinData[1], pinData[2])
					end
					if updatePins[PINS_COMPASS] and db.filters[PINS_COMPASS] then
						COMPASS_PINS.pinManager:CreatePin(PINS_COMPASS, pinData, pinData[1], pinData[2])
					end
				end
			end
		end
		
	end
	
	--if (updatePins[PINS_EIDETIC] and LMP:IsEnabled(PINS_EIDETIC)) or (updatePins[PINS_EIDETIC_COLLECTED] and LMP:IsEnabled(PINS_EIDETIC_COLLECTED)) or (updatePins[PINS_COMPASS_EIDETIC] and db.filters[PINS_COMPASS_EIDETIC]) then
	if (updatePins[PINS_EIDETIC] and LMP:IsEnabled(PINS_EIDETIC)) or (updatePins[PINS_COMPASS_EIDETIC] and db.filters[PINS_COMPASS_EIDETIC]) then
		
		local mapIndex = GetCurrentMapIndex()
		local mapContentType = GetMapContentType()
		local usePrecalculatedCoords = true
		local zoneIndex
		local eideticBooks
		
		if not mapIndex then
		
			usePrecalculatedCoords = false
			if mapContentType == MAP_CONTENT_DUNGEON then
				zoneIndex = GetCurrentMapZoneIndex()
			else
				local measurements = GPS:GetCurrentMapMeasurements()
				mapIndex = measurements.mapIndex
			end
		end
		
		if mapIndex then
			eideticBooks = LoreBooks_GetEideticDataForMap(mapIndex)
		elseif zoneIndex then
			eideticBooks = LoreBooks_GetEideticDataForZone(zoneIndex)
		end
		
		if eideticBooks then
			for _, pinData in ipairs(eideticBooks) do
				local _, _, known = GetLoreBookInfo(2, pinData.c, pinData.b)
				--if (not known and LMP:IsEnabled(PINS_EIDETIC)) or (known and LMP:IsEnabled(PINS_EIDETIC_COLLECTED)) then
				if (not known and LMP:IsEnabled(PINS_EIDETIC)) then
				
					if not pinData.qm or pinData.qm == pinData.m then
					
						if usePrecalculatedCoords and pinData.zx and pinData.zy then
							pinData.xLoc = pinData.zx
							pinData.yLoc = pinData.zy
						else
							pinData.xLoc, pinData.yLoc = GPS:GlobalToLocal(pinData.x, pinData.y)
						end
						
						local CoordsOK = pinData.xLoc and pinData.yLoc
						if CoordsOK then
							if pinData.xLoc > 0 and pinData.yLoc > 0 and pinData.xLoc < 1 and pinData.yLoc < 1 then
								if (mapContentType == MAP_CONTENT_DUNGEON and pinData.d) or mapContentType ~= MAP_CONTENT_DUNGEON then
									if not known and updatePins[PINS_EIDETIC] and LMP:IsEnabled(PINS_EIDETIC) then
										LMP:CreatePin(PINS_EIDETIC, pinData, pinData.xLoc, pinData.yLoc)
									--elseif known and updatePins[PINS_EIDETIC_COLLECTED] and LMP:IsEnabled(PINS_EIDETIC_COLLECTED) then
									--	LMP:CreatePin(PINS_EIDETIC_COLLECTED, pinData, pinData.xLoc, pinData.yLoc)
									end
								end
							end
							if updatePins[PINS_COMPASS_EIDETIC] and db.filters[PINS_COMPASS_EIDETIC] and ((mapContentType == MAP_CONTENT_DUNGEON and pinData.d) or (mapContentType ~= MAP_CONTENT_DUNGEON and pinData.d == false)) then
								COMPASS_PINS.pinManager:CreatePin(PINS_COMPASS_EIDETIC, pinData, pinData.xLoc, pinData.yLoc)
							end
						end
					end
				end
			end
			
		end
		
	end
	
	updatePins = {}
	updating = false
	
end

local function QueueCreatePins(pinType)
	updatePins[pinType] = true

	if not updating then
		updating = true
		if IsPlayerActivated() then
			if LMP.AUI.IsMinimapEnabled() then
				zo_callLater(CreatePins, 150) -- See SkyShards
			else
				CreatePins()
			end
		else
			EVENT_MANAGER:RegisterForEvent("LoreBooks_PinUpdate", EVENT_PLAYER_ACTIVATED,
				function(event)
					EVENT_MANAGER:UnregisterForEvent("LoreBooks_PinUpdate", event)
					CreatePins()
				end)
		end
	end
end

local function MapCallback_unknown()
	if not LMP:IsEnabled(PINS_UNKNOWN) or (GetMapType() > MAPTYPE_ZONE) then return end
	if GetMapContentType() == MAP_CONTENT_DUNGEON and GetCurrentZoneDungeonDifficulty() == DUNGEON_DIFFICULTY_VETERAN then return end
	QueueCreatePins(PINS_UNKNOWN)
end

local function MapCallback_collected()
	if not LMP:IsEnabled(PINS_COLLECTED) or (GetMapType() > MAPTYPE_ZONE) then return end
	if GetMapContentType() == MAP_CONTENT_DUNGEON and GetCurrentZoneDungeonDifficulty() == DUNGEON_DIFFICULTY_VETERAN then return end
	QueueCreatePins(PINS_COLLECTED)
end

local function MapCallback_eidetic()
	if not LMP:IsEnabled(PINS_EIDETIC) or (GetMapType() > MAPTYPE_ZONE) then return end
	QueueCreatePins(PINS_EIDETIC)
end

local function MapCallback_eideticCollected()
	if not LMP:IsEnabled(PINS_EIDETIC_COLLECTED) or (GetMapType() > MAPTYPE_ZONE) then return end
	QueueCreatePins(PINS_EIDETIC_COLLECTED)
end

local function CompassCallback()
	if not db.filters[PINS_COMPASS] or (GetMapType() > MAPTYPE_ZONE) then return end
	if GetMapContentType() == MAP_CONTENT_DUNGEON and GetCurrentZoneDungeonDifficulty() == DUNGEON_DIFFICULTY_VETERAN then return end
	QueueCreatePins(PINS_COMPASS)
end

local function CompassCallbackEidetic()
	if not db.filters[PINS_COMPASS_EIDETIC] or (GetMapType() > MAPTYPE_ZONE) then return end
	QueueCreatePins(PINS_COMPASS_EIDETIC)
end

-- Slash commands -------------------------------------------------------------
local function ShowMyPosition()
	if SetMapToPlayerLocation() == SET_MAP_RESULT_MAP_CHANGED then
		CALLBACK_MANAGER:FireCallbacks("OnWorldMapChanged")
	end

	local x, y = GetMapPlayerPosition("player")

	local locX = ("%05.02f"):format(zo_round(x*10000)/100)
	local locY = ("%05.02f"):format(zo_round(y*10000)/100)

	MyPrint(zo_strformat("<<1>>: <<2>>\195\151<<3>> (<<4>>)", GetMapName(), locX, locY, LMP:GetZoneAndSubzone(true)))
end

SLASH_COMMANDS["/mypos"] = ShowMyPosition
SLASH_COMMANDS["/myloc"] = ShowMyPosition

local function Base62(value)
	local r = false
	local state = type( value )
	if state == "number" then
		local k = math.floor( value )
		if k == value and value > 0 then
			local m
			r = ""
			while k > 0 do
				m = k % 62
				k = ( k - m ) / 62
				if m >= 36 then
					m = m + 61
				elseif m >= 10 then
					m = m + 55
				else
					m = m + 48
				end
				r = string.char( m ) .. r
			end
		elseif value == 0 then
			r = "0"
		end
	elseif state == "string" then
		if value:match( "^%w+$" ) then
			local n = #value
			local k = 1
			local c
			r = 0
			for i = n, 1, -1 do
				c = value:byte( i, i )
				if c >= 48 and c <= 57 then
					c = c - 48
				elseif c >= 65 and c <= 90 then
					c = c - 55
				elseif c >= 97 and c <= 122 then
					c = c - 61
				else  -- How comes?
					r = nil
					break  -- for i
				end
				r = r + c * k
				k = k * 62
			end -- for i
		end
	end
	return r
end

-- Dirty trick
local function UnsignedBase62(value)
	local isNegative = value < 0
	local value62 = Base62(math.abs(value))
	if isNegative then return "-" .. value62 end
	return value62
end

local function ConfigureMail(data)

	if data then
		
		Postmail = data
		if (not (Postmail.subject and type(Postmail.subject) == "string" and string.len(Postmail.subject) > 0)) then
			return false
		end
		if (not (Postmail.recipient and type(Postmail.recipient) == "string" and string.len(Postmail.recipient) > 0)) then
			return false
		end
		if (not (Postmail.maxDelay and type(Postmail.maxDelay) == "number" and Postmail.maxDelay >= 0)) then
			return false
		end
		if (not (Postmail.mailMaxSize and type(Postmail.mailMaxSize) == "number" and Postmail.mailMaxSize >= 0 and Postmail.mailMaxSize <= 700)) then
			return false
		end
		
		Postmail.isConfigured = true
		return true
		
	end
	
	return false

end

local function EnableMail()
	if Postmail.isConfigured then
		Postmail.isActive = true
	end
end

local function DisableMail()
	if Postmail.isConfigured and Postmail.isActive then
		Postmail.isActive = false
	end
end

local function SendData(data)

	local function SendMailData(data)
		if Postmail.recipient ~= GetDisplayName() then -- Cannot send to myself
			RequestOpenMailbox()
			SendMail(Postmail.recipient, Postmail.subject, data)
			CloseMailbox()
		else -- Directly add to POSTMAIL_DEAMON
			POSTMAIL_DEAMON[GetDisplayName() .. GetTimeStamp()] = {body = data, sender = Postmail.recipient, received = GetDate()}
		end
	end
	
	local pendingData = db.postmailData
	if Postmail.recipient == GetDisplayName() then
		SendMailData(data)
		db.postmailData = ""
	elseif Postmail.isActive then
		local dataLen = string.len(data)
		local now = GetTimeStamp()
		if pendingData ~= "" then
			if not string.find(pendingData, data) then
				local dataMergedLen = string.len(pendingData) + dataLen + 1 -- 1 is \n
				if now - db.postmailFirstInsert > Postmail.maxDelay then -- A send must be done
					if dataMergedLen > Postmail.mailMaxSize then
						SendMailData(pendingData) -- too big, send pendingData and save the modulo
						db.postmailData = data
						db.postmailFirstInsert = now
					else
						SendMailData(pendingData .. "\n" .. data) -- Send all data
						db.postmailData = ""
						db.postmailFirstInsert = now
					end
				else
					-- Send only if data is too big
					if dataMergedLen > Postmail.mailMaxSize then
						SendMailData(pendingData) -- too big, send pendingData and save the modulo
						db.postmailData = data
						db.postmailFirstInsert = now
					else
						db.postmailData = db.postmailData .. "\n" .. data
					end
				end
			end
		elseif dataLen < Postmail.mailMaxSize then
			db.postmailData = data
			db.postmailFirstInsert = now
		end
	end

end

local function BuildLorebooksLoreLibrary()

	for categoryIndex = 1, GetNumLoreCategories() do
		local categoryName, numCollections = GetLoreCategoryInfo(categoryIndex)
		for collectionIndex = 1, numCollections do
			local _, _, _, totalBooks, hidden = GetLoreCollectionInfo(categoryIndex, collectionIndex)
			if not hidden then
				for bookIndex = 1, totalBooks do
					local bookName, _, known = GetLoreBookInfo(categoryIndex, collectionIndex, bookIndex)
					if not loreBoooksLibrary[bookName] then
						loreBoooksLibrary[bookName] = {categoryIndex = categoryIndex, collectionIndex = collectionIndex, bookIndex = bookIndex}
					else
						loreBoooksLibrary[bookName].gotDuplicates = true
					end
					if known then
						totalCurrentlyCollected = totalCurrentlyCollected + 1
					end
				end
			end
		end
	end
	
end

-- Returns false if Book should not be datamined.
-- Not exact func as in PostmailDeamon cause some books still have unknown loot position and need to be datamined for check (Mainly books given by a quest)
local function EideticValidEntry(collectionIndex, bookIndex)
	
	-- Collections where Eidetic Memory mining is disabled
	
	if collectionIndex == 27 -- Dwemer
	or collectionIndex == 28 -- Akaviri
	or collectionIndex == 29 -- Glass
	or collectionIndex == 30 -- Mercenary
	or collectionIndex == 31 -- Xivkyn
	or collectionIndex == 32 -- Anc Orc
	
	or collectionIndex == 36 -- Outlaw
	or collectionIndex == 37 -- Trinimac
	or collectionIndex == 38 -- Malacath
	or collectionIndex == 39 -- Aldmeri
	or collectionIndex == 40 -- Daggerfall
	or collectionIndex == 41 -- Ebonheart
	or collectionIndex == 42 -- Dromathra
	
	or collectionIndex == 43 -- Assassin
	or collectionIndex == 44 -- Thieves Guild
	or collectionIndex == 45 -- Abah's Watch
	
	or collectionIndex == 47 -- Draugr
	or collectionIndex == 48 -- Minotaur
	or collectionIndex == 49 -- Order of the Hour
	or collectionIndex == 50 -- Hollowjack
	or collectionIndex == 51 -- Dark Brotherhood
	or collectionIndex == 52 -- Grim Harlequin
	
	then
		return false
	elseif collectionIndex == 1 then -- Base Book styles
		if bookIndex == 3 -- Hight elves
		or bookIndex == 4 -- Dark Elves
		or bookIndex == 5 -- Wood elves
		or bookIndex == 6 -- Nords
		or bookIndex == 7 -- Bretons
		or bookIndex == 8 -- Redguards
		or bookIndex == 9 -- Khajiits
		or bookIndex == 10 -- Orcs
		or bookIndex == 11 -- Argonians
		or bookIndex == 12 -- Imperial
		or bookIndex == 13 -- Anc Elves
		or bookIndex == 14 -- Barbarians
		or bookIndex == 15 -- Primal
		or bookIndex == 16 -- Daedra
		or bookIndex == 23 -- Soul-Shriven
		then
			return false
		end
	end
	
	return true
	
end

local function CoordsNearby(locX, locY, x, y)

	local nearbyIs = 0.0005 --It should be 0.00028, but we add a large diff just in case of. 0.0004 is not enough.
	if math.abs(locX - x) < nearbyIs and math.abs(locY - y) < nearbyIs then
		return true
	end
	return false
	
end

-- This code will be removed at 100015, it's only as is to thank every contributor
local function ShowThankYou(categoryIndex, collectionIndex, bookIndex)

	local icons = {
		[1] = "/esoui/art/icons/scroll_001.dds",
		[10] = "/esoui/art/icons/lore_book2_detail1_color1.dds",
		[20] = "/esoui/art/icons/quest_book_001.dds",
		[30] = "/esoui/art/icons/scroll_005.dds",
	}

	if not db.booksFound[categoryIndex] then db.booksFound[categoryIndex] = {} end
	if not db.booksFound[categoryIndex][collectionIndex] then db.booksFound[categoryIndex][collectionIndex] = {} end
	if not db.booksFound[categoryIndex][collectionIndex][bookIndex] then
		db.booksFound[categoryIndex][collectionIndex][bookIndex] = true
		db.nbBooksFound = db.nbBooksFound + 1
		if db.nbBooksFound == 1 or db.nbBooksFound == 10 or db.nbBooksFound == 20 or db.nbBooksFound == 30 then
			CENTER_SCREEN_ANNOUNCE:AddMessage(0, CSA_EVENT_COMBINED_TEXT, SOUNDS.LEVEL_UP, GetString("LBOOKS_THANK_YOU", db.nbBooksFound), GetString("LBOOKS_THANK_YOU_LONG", db.nbBooksFound), icons[db.nbBooksFound], "EsoUI/Art/Achievements/achievements_iconBG.dds")
		end
	end

end

local function BuildDataToShare(bookName, medium)
	
	if bookName and bookName ~= "" then
		
		local dataToShare
		
		local apiVersion
		if GetAPIVersion() == 100016 then
			apiVersion = 3
		else
			return
		end
		
		if SetMapToPlayerLocation() == SET_MAP_RESULT_MAP_CHANGED then
			CALLBACK_MANAGER:FireCallbacks("OnWorldMapChanged")
		end
		
		-- Will only returns the zone and not the subzone
		local zoneIndex = GetCurrentMapZoneIndex()
		
		-- Avoid nil conversion into base62
		if not zoneIndex then
			zoneIndex = 0
		end
		
		-- mapType of the subzone. Needed when we are elsewhere than zone or subzone.
		local mapContentType = GetMapContentType()
		
		local xGPS, yGPS, zoneGPS = GPS:LocalToGlobal(GetMapPlayerPosition("player"))
		
		if not zoneGPS then
			zoneGPS = 0
		end
		
		local locX = zo_round(xGPS*100000) -- 5 decimals because of Cyrodiil map
		local locY = zo_round(yGPS*100000)
		
		-- v1 = 2.4.0 LOC_DATA_UPDATE  = locX, locY, zoneIndex, mapType, lastInteractionActionWas
		-- v2 = 2.4.2 LOC_DATA_UPDATE  = locX, locY, zoneIndex, mapContentType, mapIndex, lastInteractionActionWas
		-- v3 = 2.4.4 LOC_DATA_UPDATE  = locX, locY, zoneIndex, mapContentType, mapIndex, lastInteractionActionWas, langCode
		-- v4 = 2.4.5 LOC_DATA_UPDATE  = locX, locY, zoneIndex, mapContentType, mapIndex, lastInteractionActionWas, langCode, apiVersion
		-- v5 = 2.5   LOC_DATA_UPDATE  = locX, locY, zoneIndex, mapContentType, mapIndex, lastInteractionActionWas, langCode, apiVersion, currentZoneDungeonDifficulty
		-- v6 = 2.5.1 LOC_DATA_UPDATE  = locX, locY, zoneIndex, mapContentType, mapIndex, lastInteractionActionWas, langCode, apiVersion, currentZoneDungeonDifficulty, reticleAway, associatedQuest
		-- v7 = 4     BOOK_DATA_UPDATE = categoryIndex, collectionIndex, bookIndex, mediumIndex
		dataToShare = zo_strformat("<<1>>;<<2>>;<<3>>;<<4>>;<<5>>", UnsignedBase62(locX), UnsignedBase62(locY), UnsignedBase62(zoneIndex), UnsignedBase62(mapContentType), UnsignedBase62(zoneGPS))
		
		-- if true, this is a bookshelve with a "random" book.
		if lastInteractionActionWas == GetString(LBOOKS_ACTION_SEARCH) then
			return
		else
			dataToShare = dataToShare ..";0"
		end
		
		local clang
		if lang == "en" then clang = 1 end
		if lang == "fr" then clang = 2 end
		if lang == "de" then clang = 3 end
		dataToShare = dataToShare ..";" .. clang
		
		dataToShare = dataToShare .. ";" .. apiVersion
		
		-- DungeonDifficulty for VetDungeons
		local currentZoneDungeonDifficulty = GetCurrentZoneDungeonDifficulty()
		dataToShare = dataToShare .. ";" .. currentZoneDungeonDifficulty
		
		-- We record time of lastReticle. If timer is too long we consider that the books wasn't shown by a reticle but by another way (-> inventory)
		local reticleAway = 0
		local associatedQuest = 0
		if lastReticleUpdateWas + 1500 < GetFrameTimeMilliseconds() then -- 500 is not enought
			reticleAway = 1
			
			for questIndex, questData in pairs(SHARED_INVENTORY.questCache) do
				for itemIndex, itemData in pairs(questData) do
					if string.lower(itemData.name) == string.lower(bookName) then
						associatedQuest = GetJournalQuestInfo(questIndex)
						break
					end
				end
			end
			
		end
		
		dataToShare = dataToShare .. ";" .. reticleAway .. ";" .. associatedQuest
		
		if loreBoooksLibrary[bookName] then
			
			local categoryIndex = loreBoooksLibrary[bookName].categoryIndex
			local collectionIndex = loreBoooksLibrary[bookName].collectionIndex
			local bookIndex = loreBoooksLibrary[bookName].bookIndex
			if EideticValidEntry(collectionIndex, bookIndex) then
				local bookData = LoreBooks_GetEideticData(categoryIndex, collectionIndex, bookIndex)
				if reticleAway == 0 and currentZoneDungeonDifficulty == DUNGEON_DIFFICULTY_NONE then
					if bookData and bookData.c and bookData.e then
						for _, data in ipairs(bookData.e) do
							if data.r == false and data.z == zoneIndex and CoordsNearby(xGPS, yGPS, data.x, data.y) then
								return
							end
						end
					end
				end
				
				local bookRef = zo_strformat("<<1>>;<<2>>;<<3>>", UnsignedBase62(categoryIndex), UnsignedBase62(collectionIndex), UnsignedBase62(bookIndex))
				dataToShare = dataToShare .. "@" .. bookRef
				
				dataToShare = dataToShare .. ";" .. medium
				
				if loreBoooksLibrary[bookName].gotDuplicates then
					dataToShare = dataToShare .. "@1"
				end
				
				if not bookData and categoryIndex == 2 then
					ShowThankYou(categoryIndex, collectionIndex, bookIndex)
				end
				
			else
				return
			end
			
		else
			-- Eidetic is not available until Mage Guild Questline is completed. Lorebooks PreHook can also send wrong results.
			dataToShare = dataToShare .. "@0;" .. bookName .. ";" .. medium
		end
		
		return dataToShare
		
	end
		
end
		
local function OnShowBook(_, bookName, _, medium)
	local dataToShare = BuildDataToShare(bookName, medium)
	if dataToShare then
		SendData(dataToShare)
	end
end

--Reticle Hook
local function HookTryHandlingInteraction()
	local function GetInteractionInfo(interactionPossible)
		if interactionPossible then
			if db.shareData then
				local action = GetGameCameraInteractableActionInfo()
				if action then
					lastInteractionActionWas = action
					lastReticleUpdateWas = GetFrameTimeMilliseconds()
				end
			end
		end
	end
	ZO_PreHook(RETICLE, "TryHandlingInteraction", GetInteractionInfo)
end

local function ToggleShareData()
	
	--[[
	local PostmailData = {
		subject = "LB_DATA", -- Subject of the mail
		recipient = "@Ayantir", -- Recipient of the mail. The recipient *IS GREATLY ENCOURAGED* to run the Postmail deamon
		maxDelay = 864000, -- 10d
		mailMaxSize = 600, -- Mail limitation is 700 Avoid > 650.
	}
	
	if lang == "fr" or lang == "en" or lang == "de" then
		if db.shareData then
			HookTryHandlingInteraction()
			EVENT_MANAGER:RegisterForEvent("LoreBooks", EVENT_SHOW_BOOK, OnShowBook)
			local postmailIsConfigured = ConfigureMail(PostmailData)
			if postmailIsConfigured then
				EnableMail()
			end
		else
			EVENT_MANAGER:UnregisterForEvent("LoreBooks", EVENT_SHOW_BOOK)
			DisableMail()
		end
	end
	]]--
	
end

local function ShowCongrats(newBook)
	
	local congrats = {
		[2000] = "/esoui/art/icons/quest_book_003.dds",
		[2500] = "/esoui/art/icons/scroll_001.dds",
		[2600] = "/esoui/art/icons/scroll_005.dds",
		[2700] = "/esoui/art/icons/quest_book_001.dds",
		[2800] = "/esoui/art/icons/lore_book2_detail1_color1.dds",
		-- [2821] = "/esoui/art/icons/ability_mage_064.dds", -- Max TG
		[2900] = "/esoui/art/icons/achievement_thievesguild_024.dds",
		[3000] = "/esoui/art/icons/achievement_wrothgar_027.dds",
		--[3010] = "/esoui/art/icons/ability_mage_064.dds", -- Max DB
		[3073] = "/esoui/art/icons/ability_mage_064.dds", -- Max SoH
	}
	
	local function CongratsStuff(collected)
		if not db.booksCollected[collected] then
			db.booksCollected[collected] = true
			CENTER_SCREEN_ANNOUNCE:AddMessage(0, CSA_EVENT_COMBINED_TEXT, SOUNDS.LEVEL_UP, GetString("LBOOKS_THANK_YOU", collected), GetString("LBOOKS_THANK_YOU_LONG", collected), congrats[collected], "EsoUI/Art/Achievements/achievements_iconBG.dds")
		end
	end
	
	if newBook then
		
		if congrats[totalCurrentlyCollected] then
			if (totalCurrentlyCollected == 3073 and GetAPIVersion() == 100016) then
				CongratsStuff(totalCurrentlyCollected)
			elseif totalCurrentlyCollected ~= 3073 then
				CongratsStuff(totalCurrentlyCollected)
			end
		end
		
	else
	
		-- If first activation
		EVENT_MANAGER:UnregisterForEvent("LoreBooks", EVENT_PLAYER_ACTIVATED)
		
		-- It won't show all achiev in a row, I prefer like this.
		if totalCurrentlyCollected == 3073 and GetAPIVersion() == 100016 then
			CongratsStuff(3073)
		elseif totalCurrentlyCollected >= 3000 then
			CongratsStuff(3000)
		elseif totalCurrentlyCollected >= 2900 then
			CongratsStuff(2900)
		elseif totalCurrentlyCollected >= 2800 then
			CongratsStuff(2800)
		elseif totalCurrentlyCollected >= 2700 then
			CongratsStuff(2700)
		elseif totalCurrentlyCollected >= 2600 then
			CongratsStuff(2600)
		elseif totalCurrentlyCollected >= 2500 then
			CongratsStuff(2500)
		elseif totalCurrentlyCollected >= 2000 then
			CongratsStuff(2000)
		end
		
	end
	
end

local function OnGamepadPreferredModeChanged()
	if IsInGamepadPreferredMode() then
		INFORMATION_TOOLTIP = ZO_MapLocationTooltip_Gamepad
	else
		INFORMATION_TOOLTIP = InformationTooltip
	end
end

local function OnSearchTextChanged(self)
	
	ZO_EditDefaultText_OnTextChanged(self)
	local search = self:GetText()
	LORE_LIBRARY.search = search
	
	if string.len(search) > 2 or string.len(search) == 0 then
		LORE_LIBRARY.navigationTree:ClearSelectedNode()
		LORE_LIBRARY:BuildCategoryList()
	end
	
end

local function NameSorter(left, right)
	return left.name < right.name
end

local function IsFoundInLoreLibrary(search, data)

	if string.find(string.lower(data.name), search) then
		return true
	else
		
		for bookIndex = 1, data.totalBooks do
			local title = GetLoreBookInfo(data.categoryIndex, data.collectionIndex, bookIndex)
			if string.find(string.lower(title), search) then
				return true
			end
		end
		
	end
	
	return false

end

local function ShowLoreLibraryReport(forceHide)
	
	LoreBooksCopyReport:SetHidden(true)
	if forceHide then
		reportShown = false
		LoreBooksReport:SetHidden(forceHide)
	else
		if ZO_LoreLibrary:IsHidden() then
			LoreBooksReport:SetHidden(true)
			ZO_LoreLibrary:SetHidden(false)
			reportShown = false
		else
			LoreBooksReport:SetHidden(false)
			ZO_LoreLibrary:SetHidden(true)
			reportShown = true
		end
	end
	
	KEYBIND_STRIP:UpdateKeybindButtonGroup(loreLibraryReportKeybind)
	
end

local function ShowLoreLibraryCopyReport()
	
	LoreBooksReport:SetHidden(true)
	
	LoreBooksCopyReport:GetNamedChild("Content"):GetNamedChild("Edit"):SelectAll()
	LoreBooksCopyReport:GetNamedChild("Content"):GetNamedChild("Edit"):TakeFocus()
	LoreBooksCopyReport:GetNamedChild("Content"):GetNamedChild("Edit"):SetTopLineIndex(1)
	
	LoreBooksCopyReport:SetHidden(false)
	
end

local function IsReportShown()
	return reportShown
end

local function BuildShalidorReport()

	local function DisplayCollectionsReport(collectionsData)
		
		local yCollectionIndex = 48
		
		table.sort(collectionsData, function(a,b)
			return a.totalBooks - a.numKnownBooks < b.totalBooks - b.numKnownBooks
		end)
		
		local lastObject = 0
		for collectionIndex, data in pairs(collectionsData) do
			
			local shalidorCollectionName = GetControl(LoreBooksReportContainerScrollChild, "CollectionName" .. collectionIndex)
			local shalidorCollectionValue = GetControl(LoreBooksReportContainerScrollChild, "CollectionValue" .. collectionIndex)
			
			if data.numKnownBooks ~= data.totalBooks then
				
				local shalidorCollectionName = GetControl(LoreBooksReportContainerScrollChild, "CollectionName" .. collectionIndex)
				local shalidorCollectionValue = GetControl(LoreBooksReportContainerScrollChild, "CollectionValue" .. collectionIndex)
				
				if not shalidorCollectionName then
					shalidorCollectionName = CreateControlFromVirtual("$(parent)CollectionName", LoreBooksReportContainerScrollChild, "Lorebook_ShaliCollectionName_Template", collectionIndex)
					shalidorCollectionValue = CreateControlFromVirtual("$(parent)CollectionValue", LoreBooksReportContainerScrollChild, "Lorebook_ShaliCollectionValue_Template", collectionIndex)
				end
				
				shalidorCollectionValue:SetAnchor(TOPLEFT, LoreBooksReportContainerScrollChild, TOPLEFT, 20, yCollectionIndex)
				shalidorCollectionName:SetAnchor(TOPLEFT, LoreBooksReportContainerScrollChild, TOPLEFT, 70, yCollectionIndex)
				
				yCollectionIndex = yCollectionIndex + 32
				
				shalidorCollectionName:SetText(data.name)
				shalidorCollectionValue:SetText(zo_strformat("<<1>>/<<2>>", data.numKnownBooks, data.totalBooks))
				copyReport = copyReport .. "\n\n" .. data.name  .. " :\n" ..  zo_strformat("<<1>>/<<2>>", data.numKnownBooks, data.totalBooks)
				lastObject = yCollectionIndex
			elseif shalidorCollectionName then -- Dirty trick
				shalidorCollectionName:SetHidden(true)
				shalidorCollectionValue:SetHidden(true)
			end
			
		end
		
		return lastObject + 10
		
	end

	local POINTS_FOR_RANK_MAX = 1380
	
	local totalKnown = 0
	local points = 0
	local booksInShalidor = 0
	
	collectionsData = {}
	local _, numCollections = GetLoreCategoryInfo(1)
	for collectionIndex = 1, numCollections do
		local name, _, numKnownBooks, totalBooks, hidden = GetLoreCollectionInfo(1, collectionIndex)
		if not hidden then
			totalKnown = totalKnown + numKnownBooks
			booksInShalidor = booksInShalidor + totalBooks
			points = points + numKnownBooks * 5
			collectionsData[collectionIndex] = {name = name, numKnownBooks = numKnownBooks, totalBooks = totalBooks}
			if numKnownBooks == totalBooks then
				points = points + 20
			end
		end
	end
	
	local shalidorHeaderText = GetControl(LoreBooksReport, "ShalidorHeaderText")
	local lastObject = 52
	
	if points < POINTS_FOR_RANK_MAX then
		copyReport = GetString(LBOOKS_RS_FEW_BOOKS_MISSING)
		shalidorHeaderText:SetText(copyReport)
		lastObject = DisplayCollectionsReport(collectionsData)
	elseif totalKnown < booksInShalidor then
		copyReport = GetString(LBOOKS_RS_MDONE_BOOKS_MISSING)
		shalidorHeaderText:SetText(copyReport)
		lastObject = DisplayCollectionsReport(collectionsData)
	else
		copyReport = GetString(LBOOKS_RS_GOT_ALL_BOOKS)
		shalidorHeaderText:SetText(copyReport)
	end
	
	return lastObject

end

local function BuildEideticReport(lastObject)

	local THRESHOLD_EIDETIC = 2000
	
	local eideticHeaderText = GetControl(LoreBooksReport, "EideticHeaderText")
	eideticHeaderText:ClearAnchors()
	
	eideticHeaderText:SetAnchor(TOPLEFT, LoreBooksReportContainerScrollChild, TOPLEFT, 4, lastObject)
	
	if totalCurrentlyCollected >= THRESHOLD_EIDETIC then
		
		eideticHeaderText:SetText(GetString(LBOOKS_RE_FEW_BOOKS_MISSING))
		copyReport = copyReport .. "\n\n" .. GetString(LBOOKS_RE_FEW_BOOKS_MISSING)
		
		local totalBooks = 0
		local eideticData = {}
		local eideticSeen = {}
		local yCollectionIndex = lastObject + 48
		
		local categoryName, numCollections = GetLoreCategoryInfo(2) -- Only Eidetic
		
		for collectionIndex = 1, numCollections do
			
			eideticData[collectionIndex] = {}
			eideticBooks = LoreBooks_GetEideticDataForCollection(collectionIndex)
			
			if eideticBooks then
			
				for _, bookData in ipairs(eideticBooks) do
					local _, _, known = GetLoreBookInfo(2, bookData.c, bookData.b)
					
					if not known and not eideticSeen[bookData.c .. "-" .. bookData.b] then
						table.insert(eideticData[collectionIndex], bookData)
						eideticSeen[bookData.c .. "-" .. bookData.b] = true
					end
					
				end
				
				-- Create controls
				
				local totalInCollection = #eideticData[collectionIndex]
				if totalInCollection > 0 then
				
					local eideticBooksInCollection = GetControl(LoreBooksReportContainerScrollChild, "BooksInMap" .. collectionIndex)
					local eideticCollectionName = GetControl(LoreBooksReportContainerScrollChild, "MapName" .. collectionIndex)
					local eideticReportForCollection = GetControl(LoreBooksReportContainerScrollChild, "ReportForMap" .. collectionIndex)
					
					if not eideticCollectionName then
						eideticBooksInCollection = CreateControlFromVirtual("$(parent)BooksInMap", LoreBooksReportContainerScrollChild, "Lorebook_EideticBooksInMap_Template", collectionIndex)
						eideticCollectionName = CreateControlFromVirtual("$(parent)MapName", LoreBooksReportContainerScrollChild, "Lorebook_EideticMapName_Template", collectionIndex)
						eideticReportForCollection = CreateControlFromVirtual("$(parent)ReportForMap", LoreBooksReportContainerScrollChild, "Lorebook_EideticReportForMap_Template", collectionIndex)
					end
					
					eideticBooksInCollection:SetAnchor(TOPLEFT, LoreBooksReportContainerScrollChild, TOPLEFT, 0, yCollectionIndex)
					eideticCollectionName:SetAnchor(TOPLEFT, LoreBooksReportContainerScrollChild, TOPLEFT, 25, yCollectionIndex)
					eideticReportForCollection:SetAnchor(TOPLEFT, LoreBooksReportContainerScrollChild, TOPLEFT, 50, yCollectionIndex + 24)
					
					local collectionName = LoreBooks_GetLoreCollectionName(2, collectionIndex);
					eideticBooksInCollection:SetText(totalInCollection)
					eideticCollectionName:SetText(zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, collectionName))
					
					local eideticReport = ""
					for index, data in ipairs(eideticData[collectionIndex]) do
						local bookName = GetLoreBookInfo(2, data.c, data.b)
						local bookLocation = ""
						if data.r then
							bookLocation = "[B] "
						else
							bookLocation = "[" .. GetMapNameByIndex(data.m) .. "] "
						end
						eideticReport = zo_strjoin("; ", bookLocation .. bookName, eideticReport)
					end
					
					if string.len(eideticReport) > 0 then
						eideticReport = string.sub(eideticReport, 0, -3)
					end
					
					eideticReportForCollection:SetText(eideticReport)
					copyReport = copyReport .. "\n\n" .. zo_strformat("<<1>> (<<2>>):\n<<3>>", zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, collectionName), totalInCollection, eideticReport)
					
					eideticReportForCollection:GetHeight() -- Needed to let the UI recalculate the correct value. Anchors could be optimized.
					
					yCollectionIndex = yCollectionIndex + eideticReportForCollection:GetHeight() + 32
					
					totalBooks = totalBooks + totalInCollection
					
				end
				
			end
		end
		
		copyReport = copyReport .. "\n\nTotal Remaining: " .. totalBooks
		
	else
		eideticHeaderText:SetText(GetString(LBOOKS_RE_THREESHOLD_ERROR))
		copyReport = copyReport .. "\n\n" .. GetString(LBOOKS_RE_THREESHOLD_ERROR)
	end
	
end

-- Todo : use ZO_ScrollList.
local function BuildLoreBookSummary()
	local lastObject = BuildShalidorReport()
	BuildEideticReport(lastObject)
	
	LoreBooksCopyReport:GetNamedChild("Content"):GetNamedChild("Edit"):SetText(copyReport)
	
end

local function BuildCategoryList(self)
	if self.control:IsControlHidden() then
		self.dirty = true
		return
	end
	
	self.totalCurrentlyCollected = 0
	self.totalPossibleCollected = 0
	
	self.navigationTree:Reset()
	
	local categories = {}
	
	for categoryIndex = 1, GetNumLoreCategories() do
		local categoryName, numCollections = GetLoreCategoryInfo(categoryIndex)
		for collectionIndex = 1, numCollections do
			local _, _, _, _, hidden = GetLoreCollectionInfo(categoryIndex, collectionIndex)
			if not hidden or db.unlockEidetic then
				categories[#categories + 1] = { categoryIndex = categoryIndex, name = categoryName, numCollections = numCollections }
				break
			end
		end
	end

	table.sort(categories, NameSorter)

	for i, categoryData in ipairs(categories) do
		local parent = self.navigationTree:AddNode("ZO_LabelHeader", categoryData, nil, SOUNDS.LORE_BLADE_SELECTED)

		local categoryIndex = categoryData.categoryIndex
		local numCollections = categoryData.numCollections

		local collections = {}

		for collectionIndex = 1, numCollections do
			local collectionName, description, numKnownBooks, totalBooks, hidden = GetLoreCollectionInfo(categoryIndex, collectionIndex)
			if not hidden or db.unlockEidetic then
				collections[#collections + 1] = { categoryIndex = categoryIndex, collectionIndex = collectionIndex, name = collectionName, description = description, numKnownBooks = numKnownBooks, totalBooks = totalBooks }
				self.totalCurrentlyCollected = self.totalCurrentlyCollected + numKnownBooks
				self.totalPossibleCollected = self.totalPossibleCollected + totalBooks
			end
		end

		table.sort(collections, NameSorter)
		
		local search = LORE_LIBRARY.search
		for _, collectionData in ipairs(collections) do
			if search and string.len(search) > 2 then
				if IsFoundInLoreLibrary(string.lower(search), collectionData) then
					self.navigationTree:AddNode("ZO_LoreLibraryNavigationEntry", collectionData, parent, SOUNDS.LORE_ITEM_SELECTED)
				end
			else
				self.navigationTree:AddNode("ZO_LoreLibraryNavigationEntry", collectionData, parent, SOUNDS.LORE_ITEM_SELECTED)
			end
		end
		
	end

	self.navigationTree:Commit()
	self:RefreshCollectedInfo()
	
	--Dirty hack to unselect all nodes and select the 1st one.
	
	if self.navigationTree.rootNode.children then
		if self.navigationTree.rootNode.children[1] and self.navigationTree.rootNode.children[1].children then
			self.navigationTree:SelectNode(self.navigationTree.rootNode.children[1].children[1])
		elseif self.navigationTree.rootNode.children[2] and self.navigationTree.rootNode.children[2].children then
			self.navigationTree:SelectNode(self.navigationTree.rootNode.children[2].children[1])
		end
	end
	
	KEYBIND_STRIP:UpdateKeybindButtonGroup(self.keybindStripDescriptor)
	
	self.dirty = false
	
	return true
	
end

-- "Right clic"
local function OnRowMouseUp(control, button)
	if button == MOUSE_BUTTON_INDEX_RIGHT then
		ClearMenu()
		
		-- Cannot access to self. (and self ~= control here)
		--SetMenuHiddenCallback(function() self:UnlockSelection() end)
		--self:LockSelection()
		
		if control.known then
			AddMenuItem(GetString(SI_LORE_LIBRARY_READ), function() ZO_LoreLibrary_ReadBook(control.categoryIndex, control.collectionIndex, control.bookIndex) end)
		end
		
		if IsChatSystemAvailableForCurrentPlatform() then
			AddMenuItem(GetString(SI_ITEM_ACTION_LINK_TO_CHAT), function()
				local link = ZO_LinkHandler_CreateChatLink(GetLoreBookLink, control.categoryIndex, control.collectionIndex, control.bookIndex)
				ZO_LinkHandler_InsertLink(link) 
			end)
		end
		
		if control.categoryIndex == 1 then
			local lorebookInfoOnBook = LoreBooks_GetDataOfBook(control.categoryIndex, control.collectionIndex, control.bookIndex)
			for resultEntry, resultData in ipairs(lorebookInfoOnBook) do
				
				local mapIndex = LoreBooks_GetMapIndexFromMapTile(resultData.zoneName, resultData.subZoneName)
				
				if mapIndex then
					AddMenuItem(zo_strformat("<<1>> : <<2>>x<<3>>", zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetMapNameByIndex(mapIndex)), (resultData.locX * 100), (resultData.locY * 100)),
					function()
						
						ZO_WorldMap_SetMapByIndex(mapIndex)
						PingMap(MAP_PIN_TYPE_RALLY_POINT, MAP_TYPE_LOCATION_CENTERED, resultData.locX, resultData.locY)
						PingMap(MAP_PIN_TYPE_PLAYER_WAYPOINT, MAP_TYPE_LOCATION_CENTERED, resultData.locX, resultData.locY)
						
						if(not ZO_WorldMap_IsWorldMapShowing()) then
							if IsInGamepadPreferredMode() then
								SCENE_MANAGER:Push("gamepad_worldMap")
							else
								MAIN_MENU_KEYBOARD:ShowCategory(MENU_CATEGORY_MAP)
								mapAvailable = false
							end
							mapAvailable = false
							zo_callLater(function() GPS:PanToMapPosition(resultData.locX, resultData.locY) end, 1000)
						end
						
					end)
				end
				
			end
		elseif control.categoryIndex == 2 and EideticValidEntry(control.collectionIndex, control.bookIndex) then
			
			local bookData = LoreBooks_GetEideticData(control.categoryIndex, control.collectionIndex, control.bookIndex)
			
			if bookData and bookData.c and bookData.e then
				
				if not bookData.q then
					for index, data in ipairs(bookData.e) do
					
						if not data.l and data.zx and data.zy then
						
							local xTooltip = ("%0.02f"):format(zo_round(data.zx*10000)/100)
							local yTooltip = ("%0.02f"):format(zo_round(data.zy*10000)/100)
							AddMenuItem(zo_strformat("<<1>> (<<2>>x<<3>>)", zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetMapNameByIndex(data.m)), xTooltip, yTooltip),
							function()
								
								ZO_WorldMap_SetMapByIndex(data.m)
								
								PingMap(MAP_PIN_TYPE_RALLY_POINT, MAP_TYPE_LOCATION_CENTERED, data.zx, data.zy)
								PingMap(MAP_PIN_TYPE_PLAYER_WAYPOINT, MAP_TYPE_LOCATION_CENTERED, data.zx, data.zy)
								
								if(not ZO_WorldMap_IsWorldMapShowing()) then
									if IsInGamepadPreferredMode() then
										SCENE_MANAGER:Push("gamepad_worldMap")
									else
										MAIN_MENU_KEYBOARD:ShowCategory(MENU_CATEGORY_MAP)
									end
									mapIsShowing = true
									zo_callLater(function() mapIsShowing = false end, 500) -- Bit dirty but ZO_WorldMap_IsWorldMapShowing() isn't fast enought
									zo_callLater(function() GPS:PanToMapPosition(data.zx, data.zy) end, 1000)
								end
								
							end)
							
						end
					end
				end
			end
			
		end
		
		ShowMenu(control)
		
	end
end

-- Mouse "hover"
local function OnMouseEnter(self, categoryIndex, collectionIndex, bookIndex)

	local STD_ZONE = 0

	-- No 1 for now.
	if categoryIndex == 2 and EideticValidEntry(collectionIndex, bookIndex) and (not mapIsShowing) then
	
		local bookData = LoreBooks_GetEideticData(categoryIndex, collectionIndex, bookIndex)

		if bookData and bookData.c then
			
			local bookName = GetLoreBookInfo(categoryIndex, collectionIndex, bookIndex) -- Could be retrieved automatically
			InitializeTooltip(InformationTooltip, self, BOTTOMLEFT, 0, 0, TOPRIGHT)
			InformationTooltip:AddLine(bookName, "ZoFontGameOutline", ZO_SELECTED_TEXT:UnpackRGB())
			ZO_Tooltip_AddDivider(InformationTooltip)
			
			local addDivider
			local entryWeight = {}
			
			if bookData.q then
				InformationTooltip:AddLine(GetString(LBOOKS_QUEST_BOOK), "", ZO_HIGHLIGHT_TEXT:UnpackRGB())
				InformationTooltip:AddLine(zo_strformat("[<<1>>]", bookData.q[lang] or bookData.q["en"]), "", ZO_SELECTED_TEXT:UnpackRGB())
				
				local questDetails
				if bookData.qt then
					questDetails = zo_strformat(GetString("LBOOKS_SPECIAL_QUEST"), bookData.qt)
				else
					questDetails = zo_strformat(GetString(LBOOKS_QUEST_IN_ZONE), zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetMapNameByIndex(bookData.qm)))
				end
				
				InformationTooltip:AddLine(questDetails)
				
				if bookData.h then
					InformationTooltip:AddLine(zo_strformat("[<<1>>]", GetString(LBOOKS_MAYBE_NOT_EXACT)))
				end
				
			elseif bookData.r and NonContiguousCount(bookData.m) > 1 then
				
				InformationTooltip:AddLine(GetString(LBOOKS_RANDOM_POSITION), "", ZO_HIGHLIGHT_TEXT:UnpackRGB())
				ZO_Tooltip_AddDivider(InformationTooltip)
				
				for mapIndex, count in pairs(bookData.m) do
					InformationTooltip:AddLine(zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetMapNameByIndex(mapIndex)), "", ZO_SELECTED_TEXT:UnpackRGB())
				end
				
				if bookData.h then
					InformationTooltip:AddLine(zo_strformat("[<<1>>]", GetString(LBOOKS_MAYBE_NOT_EXACT)))
				end
				
			else
				for index, data in ipairs(bookData.e) do
					
					local insert = true
					local x = data.x
					local y = data.y
					local mapIndex = data.m
					local zoneIndex = data.z
					local isRandom = data.r
					local inDungeon = data.d
					local isLost = data.l
					local specialZone = data.s
					local linkedToUnknownQuest = data.q
					
					local weight = 0
					if isRandom then
						weight = weight + 1
					end
					if inDungeon then
						weight = weight + 2
					end
					
					local zoneName = zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetZoneNameByIndex(zoneIndex))
					local mapName = zo_strformat(SI_WINDOW_TITLE_WORLD_MAP, GetMapNameByIndex(mapIndex))
					
					local bookPosition
					if zoneName ~= mapName then
						bookPosition = zo_strformat("<<1>> - <<2>>", mapName, zoneName)
						if entryWeight[bookPosition] and entryWeight[bookPosition][weight] then
							insert = false
						end
					else
						bookPosition = mapName
						if entryWeight[bookPosition] and entryWeight[bookPosition][weight] then
							insert = false
						end
					end
					
					if not entryWeight[bookPosition] then entryWeight[bookPosition] = {} end
					entryWeight[bookPosition][weight] = true
					
					if insert then
						if addDivider then
							ZO_Tooltip_AddDivider(InformationTooltip)
						end
						addDivider = true
						
						InformationTooltip:AddLine(bookPosition, "", ZO_HIGHLIGHT_TEXT:UnpackRGB())
						
						if inDungeon then
							InformationTooltip:AddLine(zo_strformat("[<<1>>]", GetString(SI_QUESTTYPE5)), "", ZO_SELECTED_TEXT:UnpackRGB())
						end
						
						if specialZone ~= STD_ZONE then
							InformationTooltip:AddLine(zo_strformat("[<<1>>]", GetString("LBOOKS_SPECIAL_ZONE", specialZone)), "", ZO_SELECTED_TEXT:UnpackRGB())
						elseif isLost then
							InformationTooltip:AddLine(zo_strformat("[<<1>>]", GetString(LBOOKS_LOST)), "", ZO_SELECTED_TEXT:UnpackRGB())
						end
						
						if isRandom then
							InformationTooltip:AddLine(GetString(LBOOKS_RANDOM_POSITION))
						end
							
					end
					
				end
				
				if bookData.h then
					InformationTooltip:AddLine(zo_strformat("[<<1>>]", GetString(LBOOKS_MAYBE_NOT_EXACT)))
				end
				
			end
		end
		
	end
	
	self.owner:EnterRow(self)

end

local function OnMouseExit(self)
	ClearTooltip(InformationTooltip)
	self.owner:ExitRow(self)
end

local function BuildBookListPostHook()
	for _, controlObject in ipairs(LORE_LIBRARY.list.list.activeControls) do
		controlObject:SetHandler("OnMouseUp", OnRowMouseUp)
		controlObject:SetHandler("OnMouseEnter", function(self) OnMouseEnter(self, controlObject.categoryIndex, controlObject.collectionIndex, controlObject.bookIndex) end)
		controlObject:SetHandler("OnMouseExit", function(self) OnMouseExit(self) end)
	end
end

local function RewriteLibrairy()

	if (lang == "fr" or lang == "en" or lang == "de") and GetAPIVersion() == 100016 then

		local original_GetNumLoreCategories = GetNumLoreCategories
		GetNumLoreCategories = function()
			if db.unlockEidetic then
				return 2
			else
				return original_GetNumLoreCategories()
			end
		end
		
		local original_GetLoreCategoryInfo = GetLoreCategoryInfo
		GetLoreCategoryInfo = function(categoryIndex)
			if db.unlockEidetic then
				return LoreBooks_GetEideticLoreCategoryInfo(categoryIndex)
			else
				return original_GetLoreCategoryInfo(categoryIndex)
			end
		end
		
		local original_GetLoreCollectionInfo = GetLoreCollectionInfo
		GetLoreCollectionInfo = function(categoryIndex, collectionIndex)
			if db.unlockEidetic then
				return LoreBooks_GetLoreCollectionInfo(categoryIndex, collectionIndex)
			else
				return original_GetLoreCollectionInfo(categoryIndex, collectionIndex)
			end
		end
	
	end
	
end

local function RebuildLoreLibrairy()
	
	loreLibraryReportKeybind =
	{
		{
			alignment = KEYBIND_STRIP_ALIGN_LEFT,
			name = "Report",
			keybind = "UI_SHORTCUT_SECONDARY",
			callback = ShowLoreLibraryReport,
		},
		{
			alignment = KEYBIND_STRIP_ALIGN_LEFT,
			name = "Copy",
			keybind = "UI_SHORTCUT_TERTIARY",
			callback = ShowLoreLibraryCopyReport,
			visible = IsReportShown,
		},
	}
	
	local function OnStateChanged(oldState, newState)
		if newState == SCENE_SHOWING then
			KEYBIND_STRIP:AddKeybindButtonGroup(loreLibraryReportKeybind)
		elseif newState == SCENE_HIDDEN then
			KEYBIND_STRIP:RemoveKeybindButtonGroup(loreLibraryReportKeybind)
			ShowLoreLibraryReport(true)
		end
	end
	
	LORE_LIBRARY_SCENE:RegisterCallback("StateChange", OnStateChanged)
	
	local lorebookResearch = WINDOW_MANAGER:CreateControlFromVirtual("Lorebook_Research", ZO_LoreLibrary, "Lorebook_Research_Template")
	lorebookResearch.searchBox = GetControl(lorebookResearch, "Box")
	lorebookResearch.searchBox:SetHandler("OnTextChanged", OnSearchTextChanged)
	
	ZO_PreHook(LORE_LIBRARY, "BuildCategoryList", BuildCategoryList)
	
	RewriteLibrairy()
	BuildLorebooksLoreLibrary()
	BuildLoreBookSummary()
	
	local origLoreLibraryBuildBookList = LORE_LIBRARY.BuildBookList
	LORE_LIBRARY.BuildBookList = function(self, ...)
		origLoreLibraryBuildBookList(self, ...)
		BuildBookListPostHook()
	end
	
end

local function CreatePins()

	local pinTextureLevel = db.pinTexture.level
	local pinTextureSize = db.pinTexture.size
	local compassMaxDistance = db.compassMaxDistance
	
	local mapPinLayout_eidetic = { level = pinTextureLevel, texture = GetPinTextureEidetic, size = pinTextureSize }
	local mapPinLayout_eideticCollected = { level = pinTextureLevel, texture = GetPinTextureEidetic, size = pinTextureSize }
	local mapPinLayout_unknown = { level = pinTextureLevel, texture = GetPinTexture, size = pinTextureSize }
	local mapPinLayout_collected = { level = pinTextureLevel, texture = GetPinTexture, size = pinTextureSize, grayscale = IsPinGrayscale }
	local compassPinLayout = { maxDistance = compassMaxDistance, texture = pinTextures[db.pinTexture.type][2],
		sizeCallback = function(pin, angle, normalizedAngle, normalizedDistance)
			if zo_abs(normalizedAngle) > 0.25 then
				pin:SetDimensions(54 - 24 * zo_abs(normalizedAngle), 54 - 24 * zo_abs(normalizedAngle))
			else
				pin:SetDimensions(48, 48)
			end
		end,
		additionalLayout = {
			function(pin, angle, normalizedAngle, normalizedDistance)
				if (pinTexturesList[db.pinTexture.type] == "Shalidor's Library icons") then --replace icon with icon from LoreLibrary
					local _, texture = GetLoreBookInfo(1, pin.pinTag[3], pin.pinTag[4])
					if icon == MISSING_TEXTURE then icon = PLACEHOLDER_TEXTURE end
					local icon = pin:GetNamedChild("Background")
					icon:SetTexture(texture)
				end
			end,
			function(pin)
				--I do not need to reset anything (texture is changed automatically), so the function is empty
			end
		}
	}
	local compassPinLayoutEidetic = { maxDistance = compassMaxDistance, texture = pinTextures[db.pinTexture.type][2],
		sizeCallback = function(pin, angle, normalizedAngle, normalizedDistance)
			if zo_abs(normalizedAngle) > 0.25 then
				pin:SetDimensions(54 - 24 * zo_abs(normalizedAngle), 54 - 24 * zo_abs(normalizedAngle))
			else
				pin:SetDimensions(48, 48)
			end
		end,
		additionalLayout = {
			function(pin, angle, normalizedAngle, normalizedDistance)
				if (pinTexturesList[db.pinTexture.type] == "Shalidor's Library icons") then --replace icon with icon from LoreLibrary
					local _, texture = GetLoreBookInfo(2, pin.pinTag.c, pin.pinTag.b)
					if icon == MISSING_TEXTURE then icon = PLACEHOLDER_TEXTURE end
					local icon = pin:GetNamedChild("Background")
					icon:SetTexture(texture)
				end
			end,
			function(pin)
				--I do not need to reset anything (texture is changed automatically), so the function is empty
			end
		}
	}
	
	--initialize map pins
	LMP:AddPinType(PINS_UNKNOWN, MapCallback_unknown, nil, mapPinLayout_unknown, pinTooltipCreator)
	LMP:AddPinType(PINS_COLLECTED, MapCallback_collected, nil, mapPinLayout_collected, pinTooltipCreator)
	LMP:AddPinType(PINS_EIDETIC, MapCallback_eidetic, nil, mapPinLayout_eidetic, pinTooltipCreatorEidetic)
	--LMP:AddPinType(PINS_EIDETIC_COLLECTED, MapCallback_eideticCollected, nil, mapPinLayout_eideticCollected, pinTooltipCreatorEidetic)
	
	--add map filters
	LMP:AddPinFilter(PINS_UNKNOWN, GetString(LBOOKS_FILTER_UNKNOWN), nil, db.filters)
	LMP:AddPinFilter(PINS_COLLECTED, GetString(LBOOKS_FILTER_COLLECTED), nil, db.filters)
	LMP:AddPinFilter(PINS_EIDETIC, GetLoreCategoryInfo(2), nil, db.filters)
	--LMP:AddPinFilter(PINS_EIDETIC_COLLECTED, GetLoreCategoryInfo(2) .. " (collecté)", nil, db.filters)
	
	--add handler for the left click
	LMP:SetClickHandlers(PINS_UNKNOWN, {
		[1] = {
			name = function(pin) return zo_strformat(LBOOKS_SET_WAYPOINT, GetLoreBookInfo(1, pin.m_PinTag[3], pin.m_PinTag[4])) end,
			show = function(pin) return not select(3, GetLoreBookInfo(1, pin.m_PinTag[3], pin.m_PinTag[4])) end,
			duplicates = function(pin1, pin2) return (pin1.m_PinTag[3] == pin2.m_PinTag[3] and pin1.m_PinTag[4] == pin2.m_PinTag[4]) end,
			callback = function(pin) PingMap(MAP_PIN_TYPE_PLAYER_WAYPOINT, MAP_TYPE_LOCATION_CENTERED, pin.normalizedX, pin.normalizedY) end,
		}
	})
	
	--initialize compass pins
	COMPASS_PINS:AddCustomPin(PINS_COMPASS, CompassCallback, compassPinLayout)
	COMPASS_PINS:RefreshPins(PINS_COMPASS)
	
	COMPASS_PINS:AddCustomPin(PINS_COMPASS_EIDETIC, CompassCallbackEidetic, compassPinLayoutEidetic)
	COMPASS_PINS:RefreshPins(PINS_COMPASS_EIDETIC)
	
end

local function OnBookLearned(eventCode, categoryIndex, collectionIndex, bookIndex)
	
	totalCurrentlyCollected = totalCurrentlyCollected + 1
	
	--Refresh map if needed and get player position
	if SetMapToPlayerLocation() == SET_MAP_RESULT_MAP_CHANGED then
		CALLBACK_MANAGER:FireCallbacks("OnWorldMapChanged")
	end
	local x, y = GetMapPlayerPosition("player")

	if not InvalidPoint(x, y) then -- can be false in some very rare place (mainly fighters/mages/main questline dungeons).
		LMP:RefreshPins(PINS_UNKNOWN)
		LMP:RefreshPins(PINS_COLLECTED)
		LMP:RefreshPins(PINS_EIDETIC)
		--LMP:RefreshPins(PINS_EIDETIC_COLLECTED)
		COMPASS_PINS:RefreshPins(PINS_COMPASS)
		COMPASS_PINS:RefreshPins(PINS_COMPASS_EIDETIC)
	end
	
	BuildLoreBookSummary()
	
	-- LORE_LIBRARY need to be refreshed first
	zo_callLater(function() ShowCongrats(true) end, 50)
	
end

-- Settings menu --------------------------------------------------------------
local function CreateSettingsMenu()
	local panelData = {
		type = "panel",
		name = GetString(LBOOKS_TITLE),
		displayName = ZO_HIGHLIGHT_TEXT:Colorize(GetString(LBOOKS_TITLE)),
		author = "Garkin & Ayantir",
		version = ADDON_VERSION,
		slashCommand = "/lorebooks",
		registerForRefresh = true,
		registerForDefaults = true,
	}
	LAM:RegisterAddonPanel("LoreBooks", panelData)

	local CreateIcons, unknownIcon, collectedIcon
	CreateIcons = function(panel)
		if panel == LoreBooks then
			unknownIcon = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[1], CT_TEXTURE)
			unknownIcon:SetAnchor(RIGHT, panel.controlsToRefresh[1].combobox, LEFT, -10, 0)
			unknownIcon:SetTexture(pinTextures[db.pinTexture.type][2])
			unknownIcon:SetDimensions(db.pinTexture.size, db.pinTexture.size)
			collectedIcon = WINDOW_MANAGER:CreateControl(nil, panel.controlsToRefresh[1], CT_TEXTURE)
			collectedIcon:SetAnchor(RIGHT, unknownIcon, LEFT, -5, 0)
			collectedIcon:SetTexture(pinTextures[db.pinTexture.type][1])
			collectedIcon:SetDimensions(db.pinTexture.size, db.pinTexture.size)
			collectedIcon:SetDesaturation((pinTexturesList[db.pinTexture.type] == "Shalidor's Library icons") and 1 or 0)
			CALLBACK_MANAGER:UnregisterCallback("LAM-PanelControlsCreated", CreateIcons)
		end
	end
	CALLBACK_MANAGER:RegisterCallback("LAM-PanelControlsCreated", CreateIcons)

	local optionsTable = {
		{
			type = "dropdown",
			name = GetString(LBOOKS_PIN_TEXTURE),
			tooltip = GetString(LBOOKS_PIN_TEXTURE_DESC),
			choices = pinTexturesList,
			getFunc = function() return pinTexturesList[db.pinTexture.type] end,
			setFunc = function(selected)
					for index, name in ipairs(pinTexturesList) do
						if name == selected then
							db.pinTexture.type = index
							unknownIcon:SetTexture(pinTextures[index][2])
							collectedIcon:SetDesaturation((pinTexturesList[index] == "Shalidor's Library icons") and 1 or 0)
							collectedIcon:SetTexture(pinTextures[index][1])
							LMP:RefreshPins(PINS_UNKNOWN)
							LMP:RefreshPins(PINS_COLLECTED)
							LMP:RefreshPins(PINS_EIDETIC)
							--LMP:RefreshPins(PINS_EIDETIC_COLLECTED)
							COMPASS_PINS.pinLayouts[PINS_COMPASS].texture = pinTextures[index][2]
							COMPASS_PINS:RefreshPins(PINS_COMPASS)
							COMPASS_PINS.pinLayouts[PINS_COMPASS_EIDETIC].texture = pinTextures[index][2]
							COMPASS_PINS:RefreshPins(PINS_COMPASS_EIDETIC)
							break
						end
					end
				end,
			--disabled = function() return not (db.filters[PINS_UNKNOWN] or db.filters[PINS_COLLECTED] or db.filters[PINS_EIDETIC] or db.filters[PINS_EIDETIC_COLLECTED]) end,
			disabled = function() return not (db.filters[PINS_UNKNOWN] or db.filters[PINS_COLLECTED] or db.filters[PINS_EIDETIC]) end,
			default = pinTexturesList[defaults.pinTexture.type],
		},
		{
			type = "slider",
			name = GetString(LBOOKS_PIN_SIZE),
			tooltip = GetString(LBOOKS_PIN_SIZE_DESC),
			min = 20,
			max = 70,
			step = 1,
			getFunc = function() return db.pinTexture.size end,
			setFunc = function(size)
					db.pinTexture.size = size
					unknownIcon:SetDimensions(size, size)
					collectedIcon:SetDimensions(size, size)
					LMP:SetLayoutKey(PINS_UNKNOWN, "size", size)
					LMP:SetLayoutKey(PINS_COLLECTED, "size", size)
					LMP:SetLayoutKey(PINS_EIDETIC, "size", size)
					--LMP:SetLayoutKey(PINS_EIDETIC_COLLECTED, "size", size)
					LMP:RefreshPins(PINS_UNKNOWN)
					LMP:RefreshPins(PINS_COLLECTED)
					LMP:RefreshPins(PINS_EIDETIC)
					--LMP:RefreshPins(PINS_EIDETIC_COLLECTED)
				end,
			--disabled = function() return not (db.filters[PINS_UNKNOWN] or db.filters[PINS_COLLECTED] or db.filters[PINS_EIDETIC] or db.filters[PINS_EIDETIC_COLLECTED]) end,
			disabled = function() return not (db.filters[PINS_UNKNOWN] or db.filters[PINS_COLLECTED] or db.filters[PINS_EIDETIC]) end,
			default = defaults.pinTexture.size
		},
		{
			type = "slider",
			name = GetString(LBOOKS_PIN_LAYER),
			tooltip = GetString(LBOOKS_PIN_LAYER_DESC),
			min = 10,
			max = 200,
			step = 5,
			getFunc = function() return db.pinTexture.level end,
			setFunc = function(level)
					db.pinTexture.level = level
					LMP:SetLayoutKey(PINS_UNKNOWN, "level", level)
					LMP:SetLayoutKey(PINS_COLLECTED, "level", level)
					LMP:SetLayoutKey(PINS_EIDETIC, "level", level)
					--LMP:SetLayoutKey(PINS_EIDETIC_COLLECTED, "level", level)
					LMP:RefreshPins(PINS_UNKNOWN)
					LMP:RefreshPins(PINS_COLLECTED)
					LMP:RefreshPins(PINS_EIDETIC)
					--LMP:RefreshPins(PINS_EIDETIC_COLLECTED)
				end,
			--disabled = function() return not (db.filters[PINS_UNKNOWN] or db.filters[PINS_COLLECTED] or db.filters[PINS_EIDETIC] or db.filters[PINS_EIDETIC_COLLECTED]) end,
			disabled = function() return not (db.filters[PINS_UNKNOWN] or db.filters[PINS_COLLECTED] or db.filters[PINS_EIDETIC]) end,
			default = defaults.pinTexture.level,
		},
		{
			type = "checkbox",
			name = GetString(LBOOKS_UNKNOWN),
			tooltip = GetString(LBOOKS_UNKNOWN_DESC),
			getFunc = function() return db.filters[PINS_UNKNOWN] end,
			setFunc = function(state)
					db.filters[PINS_UNKNOWN] = state
					LMP:SetEnabled(PINS_UNKNOWN, state)
				end,
			default = defaults.filters[PINS_UNKNOWN],
		},
		{
			type = "checkbox",
			name = GetString(LBOOKS_COLLECTED),
			tooltip = GetString(LBOOKS_COLLECTED_DESC),
			getFunc = function() return db.filters[PINS_COLLECTED] end,
			setFunc = function(state)
					db.filters[PINS_COLLECTED] = state
					LMP:SetEnabled(PINS_COLLECTED, state)
				end,
			default = defaults.filters[PINS_COLLECTED]
		},
		{
			type = "checkbox",
			name = GetString(LBOOKS_EIDETIC),
			tooltip = GetString(LBOOKS_EIDETIC_DESC),
			getFunc = function() return db.filters[PINS_EIDETIC] end,
			setFunc = function(state)
					db.filters[PINS_EIDETIC] = state
					LMP:SetEnabled(PINS_EIDETIC, state)
				end,
			default = defaults.filters[PINS_EIDETIC]
		},
		--[[
		{
			type = "checkbox",
			name = GetString(LBOOKS_EIDETIC),
			tooltip = GetString(LBOOKS_EIDETIC_DESC),
			getFunc = function() return db.filters[PINS_EIDETIC_COLLECTED] end,
			setFunc = function(state)
					db.filters[PINS_EIDETIC_COLLECTED] = state
					LMP:SetEnabled(PINS_EIDETIC_COLLECTED, state)
				end,
			default = defaults.filters[PINS_EIDETIC_COLLECTED]
		},
		]]
		{
			type = "checkbox",
			name = GetString(LBOOKS_COMPASS_UNKNOWN),
			tooltip = GetString(LBOOKS_COMPASS_UNKNOWN_DESC),
			getFunc = function() return db.filters[PINS_COMPASS] end,
			setFunc = function(state)
					db.filters[PINS_COMPASS] = state
					COMPASS_PINS:RefreshPins(PINS_COMPASS)
				end,
			default = defaults.filters[PINS_COMPASS],
		},
		{
			type = "checkbox",
			name = GetString(LBOOKS_COMPASS_EIDETIC),
			tooltip = GetString(LBOOKS_COMPASS_EIDETIC_DESC),
			getFunc = function() return db.filters[PINS_COMPASS_EIDETIC] end,
			setFunc = function(state)
					db.filters[PINS_COMPASS_EIDETIC] = state
					COMPASS_PINS:RefreshPins(PINS_COMPASS_EIDETIC)
				end,
			default = defaults.filters[PINS_COMPASS_EIDETIC],
		},
		{
			type = "slider",
			name = GetString(LBOOKS_COMPASS_DIST),
			tooltip = GetString(LBOOKS_COMPASS_DIST_DESC),
			min = 1,
			max = 100,
			step = 1,
			getFunc = function() return db.compassMaxDistance * 1000 end,
			setFunc = function(maxDistance)
					db.compassMaxDistance = maxDistance / 1000
					COMPASS_PINS.pinLayouts[PINS_COMPASS].maxDistance = maxDistance / 1000
					COMPASS_PINS:RefreshPins(PINS_COMPASS)
					COMPASS_PINS.pinLayouts[PINS_COMPASS_EIDETIC].maxDistance = maxDistance / 1000
					COMPASS_PINS:RefreshPins(PINS_COMPASS_EIDETIC)
				end,
			disabled = function() return not (db.filters[PINS_COMPASS] or db.filters[PINS_COMPASS_EIDETIC]) end,
			default = defaults.compassMaxDistance * 1000,
		},
		{
			type = "checkbox",
			name = GetString(LBOOKS_UNLOCK_EIDETIC),
			tooltip = GetString(LBOOKS_UNLOCK_EIDETIC_DESC),
			getFunc = function() return db.unlockEidetic end,
			setFunc = function(state)
				db.unlockEidetic = state
				LORE_LIBRARY:BuildCategoryList()
			end,
			default = defaults.unlockEidetic,
			disabled = function() return GetAPIVersion() ~= 100016 end,
		},
		{
			type = "checkbox",
			name = GetString(LBOOKS_SHARE_DATA),
			tooltip = GetString(LBOOKS_SHARE_DATA_DESC),
			getFunc = function() return db.shareData end,
			setFunc = function(state)
				db.shareData = state
				ToggleShareData()
				end,
			default = defaults.shareData,
			disabled = true, --(GetWorldName() ~= "EU Megaserver" or not (lang == "fr" or lang == "en" or lang == "de"))
		},
	}
	LAM:RegisterOptionControls("LoreBooks", optionsTable)
end

local function OnLoad(eventCode, name)

	if name == "LoreBooks" then
		EVENT_MANAGER:UnregisterForEvent("LoreBooks", EVENT_ADD_ON_LOADED)
		
		--saved variables for settings
		db = ZO_SavedVars:NewCharacterNameSettings("LBooks_SavedVariables", 2, nil, defaults)
		
		-- Settings
		CreateSettingsMenu()
		
		-- Lorelibrary
		RebuildLoreLibrairy()
		
		-- Tooltip Mode
		OnGamepadPreferredModeChanged()
		
		-- LibMapPins
		CreatePins()

		-- Data sniffer
		ToggleShareData()
		
		--events
		EVENT_MANAGER:RegisterForEvent("LoreBooks", EVENT_LORE_BOOK_LEARNED, OnBookLearned)
		EVENT_MANAGER:RegisterForEvent("LoreBooks", EVENT_PLAYER_ACTIVATED, ShowCongrats)
		EVENT_MANAGER:RegisterForEvent("LoreBooks", EVENT_GAMEPAD_PREFERRED_MODE_CHANGED, OnGamepadPreferredModeChanged)
		
	end
	
end

EVENT_MANAGER:RegisterForEvent("LoreBooks", EVENT_ADD_ON_LOADED, OnLoad)